"""Authentication helper class."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Optional


class AuthToken:
    """Authenticate token helper class."""

    def __init__(
        self,
        refresh_token: Optional[str] = "",
        refresh_ts: Optional[datetime] = None,
        token: Optional[str] = "",
        user_id: Optional[str] = "",
    ) -> None:
        """Initialize."""
        self._token = token
        self._refresh_ts = refresh_ts
        self._refresh_token = refresh_token
        self._user_id = user_id

    @property
    def token(self):
        """Return active token."""
        return self._token

    @property
    def refresh_token(self):
        """Return refresh token."""
        return self._refresh_token

    @property
    def refresh_timestamp(self):
        """Return timestmap of when token needs to be refreshed."""
        return self._refresh_ts

    @property
    def user_id(self):
        """Return user id."""
        return self._user_id

    @property
    def is_valid(self):
        """Return true if token is valid."""
        if not self._refresh_ts:
            return False

        return datetime.now() < self._refresh_ts

    def __repr__(self):
        """Return repr of object."""
        return "AuthToken(%s, %s, %s, %s)" % (
            self._refresh_token.__repr__(),
            self._refresh_ts.__repr__(),
            self._token.__repr__(),
            self._user_id.__repr__(),
        )



"""Define a class to hold all zones."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Awaitable, Callable, Dict

from .zone import Zone


class Zones:
    """Represents all zones in a controller."""

    def __init__(self, request: Callable[..., Awaitable[dict]], zones: list):
        """Initialize."""
        self._request = request
        self._zones = {}  # type: Dict[str, Any]
        for item in zones:
            zone = Zone(item, request)
            self._zones[zone.id] = zone

    def __iter__(self):
        """Iterator."""
        for item in self._zones:
            yield self._zones[item]

    def __getitem__(self, key):
        """Iterator."""
        if key in self._zones:
            return self._zones[key]
        raise KeyError

    def __len__(self):
        """Return number of zones."""
        return len(self._zones)

    def get(self, idx: str):
        """Get zone by id."""
        return self._zones.get(idx)

    def get_by_number(self, number: int):
        """Get zone by number."""
        for pos, (key, zone) in enumerate(self._zones.items()):
            if zone.number == number:
                return zone

        raise KeyError()

    # pylint: disable=unused-variable
    async def run(self, run_zones: list):
        """Run zones."""
        run_list = []
        for item in run_zones:
            if not isinstance(item, dict):
                raise Exception("run takes a list of dict(zone-number, time)")

            if len(item) != 1:
                raise Exception("run takes a list of dict(zone-number, time)")

            for pos, (zone, time) in enumerate(item.items()):
                run_list.append({"zone": zone, "time": time})

        return await self._request("post", "run", json=run_list)

    async def skip(self):
        """Skip the current running zone and go to the next in the queue."""
        return await self._request("post", "skip")



"""Define class to handle webhooks."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Awaitable, Callable, List

from .errors import InvalidWebhookEvent
from .dataobject import ListObject
from .webhook import Webhook


class Webhooks:
    """Handle webhooks for Sprinkl cloud service."""

    def __init__(self, request: Callable[..., Awaitable[dict]]) -> None:
        """Initialize."""
        self._request = request
        self._events = None

    async def create(self, external_id: str, url: str, events: List[str]) -> Webhook:
        """Create a new webhook for this controller."""
        supported_events = await self.events()
        for event in events:
            if event not in supported_events:
                raise InvalidWebhookEvent()

        wh_create = {"external_id": external_id, "url": url, "events": events}

        create = await self._request("post", "webhooks", json=wh_create)
        return Webhook(create["data"], self._request)

    async def events(self) -> List[str]:
        """List all webhook events."""
        if self._events:
            return self._events

        events = await self._request("get", "webhooks/events")
        self._events = events["data"]
        assert self._events is not None
        return self._events

    async def get(self, ident: str) -> Webhook:
        """Get webhook by id."""
        webhook = await self._request("get", "webhooks/{0}".format(ident))
        return Webhook(webhook["data"], self._request)

    async def delete(self, ident: str) -> None:
        """Delete webhook by id."""
        await self._request("delete", "webhooks/{0}".format(ident))

    async def all(self) -> ListObject:
        """Return all active webhooks."""
        webhooks = await self._request("get", "webhooks")
        return ListObject([Webhook(item, self._request) for item in webhooks["data"]])

    async def find(self, external_id: str) -> ListObject:
        """Find all webhooks matching a external id."""
        webhooks = await self.all()
        hooks = [webhook for webhook in webhooks if webhook.external_id == external_id]
        return ListObject(hooks)



"""Client class for Sprinkl controllers."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

import async_timeout
from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientResponseError

from .authtoken import AuthToken
from .const import (
    DEFAULT_TIMEOUT,
    SPRINKL_AUTH_ENDPOINT,
    SPRINKL_ENDPOINT,
    TOKEN_LIFETIME,
    USER_AGENT,
)
from .controller import Controller
from .errors import (
    AuthenticateError,
    ControllerAlreadyRunning,
    RequestError,
    RequestTimeout,
    TokenExpired,
)

_LOGGER = logging.getLogger(__name__)


class Client:
    """Client class for Sprinkl controllers."""

    def __init__(
        self,
        websession: ClientSession,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
        ssl: Optional[bool] = True,
        proxy: Optional[str] = None,
    ) -> None:
        """Initialize."""
        self._websession = websession
        self._timeout = timeout
        self._ssl = ssl
        self._proxy = proxy
        self._auth = None  # type: Union[AuthToken, None]

    async def _request(
        self,
        method: str,
        url: str,
        *,
        authorization: Optional[str] = None,
        headers: Optional[dict] = None,
        params: Optional[dict] = None,
        json: Optional[dict] = None,
        reauth_token: Optional[bool] = False
    ) -> Dict[Any, Any]:
        if not headers:
            headers = {}
        if not params:
            params = {}

        headers.update({"User-Agent": USER_AGENT, "Content-Type": "application/json"})

        if authorization:
            headers.update({"Authorization": authorization})

        if not authorization and self._auth:
            headers.update({"Authorization": self._auth.token})

        try:
            while True:
                try:
                    async with async_timeout.timeout(self._timeout):
                        async with self._websession.request(
                            method,
                            url,
                            headers=headers,
                            params=params,
                            json=json,
                            ssl=self._ssl,
                            proxy=self._proxy,
                        ) as response:
                            await _throw_api_exception(response)
                            response.raise_for_status()
                            return await response.json(content_type=None)
                except TokenExpired as err:
                    if not self._auth or not reauth_token:
                        if self._auth is None:
                            _LOGGER.error("Reqest to re-auth but no auth token.")
                        raise err
                    # Try re-auth only once
                    reauth_token = False
                    self._auth = await self._try_refresh_token_auth(self._auth)
                    if self._auth is None:
                        _LOGGER.error("Failed to refresh token while doing re-auth.")
                except ClientResponseError as err:
                    _LOGGER.error("Request error: %s (%s)", err, type(err))
                    raise RequestError(err)
        except asyncio.TimeoutError as err:
            raise RequestTimeout(err)

    async def _try_refresh_token_auth(
        self, auth_info: AuthToken
    ) -> Union[AuthToken, None]:
        try:
            auth = await self._request(
                "post",
                SPRINKL_AUTH_ENDPOINT,
                params={"refresh_token": auth_info.refresh_token},
            )
            return _create_auth_info(auth)
        except AuthenticateError as err:
            _LOGGER.error(
                "Failed to authenticate while getting refresh token. (%s)", str(err)
            )
            return None

    async def _try_login(
        self, email: Optional[str], password: Optional[str]
    ) -> AuthToken:
        if email is None or password is None:
            raise AuthenticateError("Failed to login with credentials")

        auth = await self._request(
            "post", SPRINKL_AUTH_ENDPOINT, json={"email": email, "password": password}
        )
        return _create_auth_info(auth)

    async def _get_controllers(self, auth_info: AuthToken) -> List[Any]:
        controllers = await self._request(
            "get", SPRINKL_ENDPOINT + "/controllers", authorization=auth_info.token
        )
        return controllers["data"]

    def _parse_controllers(self, controllers: list) -> dict:
        parsed = {}
        for controller in controllers:
            obj = Controller(self._request, controller)
            parsed[obj.id] = obj
        return parsed

    @property
    def auth_info(self):
        """Return active authentication information."""
        return self._auth

    # pylint: disable=attribute-defined-outside-init
    async def login(
        self, email: str = None, password: str = None, auth_info: AuthToken = None
    ) -> Union[AuthToken, None]:
        """Login to Sprinkl cloud and get all controllers."""
        controllers = None
        if auth_info:
            if auth_info.is_valid:
                try:
                    controllers = await self._get_controllers(auth_info)
                except TokenExpired:
                    # Try using refresh token
                    if auth_info.refresh_token:
                        auth_info = await self._try_refresh_token_auth(auth_info)
                    else:
                        auth_info = None
                except AuthenticateError:
                    # just continue and try to login
                    auth_info = None
                    pass
            else:
                # Try using refresh token
                if auth_info.refresh_token:
                    auth_info = await self._try_refresh_token_auth(auth_info)
                else:
                    auth_info = None

        # Try to login if refresh / controller failed
        if not auth_info:
            auth_info = await self._try_login(email, password)

        if not controllers:
            controllers = await self._get_controllers(auth_info)

        self._controllers = self._parse_controllers(controllers)
        self._auth = auth_info

        return auth_info

    async def controllers(self) -> list:
        """Return controllers."""
        return [self._controllers[key] for key in self._controllers]

    async def get(self, idx: str) -> Controller:
        """Get controller by index."""
        return self._controllers[idx]


def _create_auth_info(auth: dict) -> AuthToken:
    return AuthToken(
        token=auth["data"]["token"],
        refresh_token=auth["data"]["refresh_token"],
        refresh_ts=datetime.now() + TOKEN_LIFETIME,
        user_id=auth["data"]["user_id"],
    )


# pylint: disable=broad-except
async def _get_response_json(response) -> Any:
    try:
        return await response.json(content_type=None)
    except Exception as err:
        _LOGGER.info(
            "_get_response_json() Failed to parse json: %s (%s)", err, type(err)
        )
        # broken json for some errors
        pass
    return None


async def _throw_api_exception(response) -> None:
    if response.status == 403:
        json = await _get_response_json(response)
        if (
            json
            and "errors" in json
            and len(json["errors"]) == 1
            and json["errors"][0]["detail"] == "Controller is already running"
        ):
            raise ControllerAlreadyRunning()

    if response.status == 401:
        # TODO(totte): not sure this is a good idea - maybe just always try to refresh token
        json = await _get_response_json(response)
        if (
            json
            and "errors" in json
            and len(json["errors"]) == 1
            and json["errors"][0]["detail"] == "Token has expired"
        ):
            raise TokenExpired()

        raise AuthenticateError()



"""Define class to handle a webhook."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Awaitable, Callable

from .dataobject import DictObject


class Webhook(DictObject):
    """Handle webhooks for Sprinkl cloud service."""

    def __init__(self, webhook, request: Callable[..., Awaitable[dict]]):
        """Initialize."""
        super().__init__(webhook)
        self._request = request

    async def delete(self) -> None:
        """Delete current webhook."""
        await self._request("delete", "webhooks/{0}".format(self.id))



"""Define a zone class."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Awaitable, Callable

from .dataobject import DictObject


class Zone(DictObject):
    """Represents a zone in the Sprinkl controller."""

    def __init__(self, data: dict, request: Callable[..., Awaitable[dict]]):
        """Initialize."""
        super().__init__(data)
        self._request = request

    @property
    def id(self) -> str:
        """Return id of zone."""
        return self.get("id")

    @property
    def number(self) -> int:
        """Return numer of zone."""
        return self.get("number")

    @property
    def enabled(self) -> bool:
        """Return if the zone is enabled."""
        return self.get("enabled")

    async def run(self, time_in_minutes: int) -> None:
        """Run the current zone."""
        await self._request(
            "post", "run", json=[{"zone": self.number, "time": time_in_minutes}]
        )



"""Define schedule class."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import date
from typing import Awaitable, Callable

from .dataobject import DictObject


class Schedule(DictObject):
    """Defines a schedule."""

    def __init__(
        self, request: Callable[..., Awaitable[dict]], adjustments: dict, schedule: dict
    ):
        """Intialize."""
        super().__init__(schedule)
        self._request = request
        self._adjustments = adjustments

    @property
    def enabled(self) -> bool:
        """Return true if sechedule is enabled."""
        return self.get("enabled")

    async def run(self, use_seasonal_adjustment: bool = False) -> None:
        """Run a schedule manually with adjustments if needed."""
        zones_to_run = []

        for zone in self.zones:
            run_time = zone.run_time
            if use_seasonal_adjustment:
                run_time = round(
                    run_time * (self._adjustments[date.today().month - 1] / 100)
                )

            zones_to_run.append({"zone": zone.number, "time": run_time})

        await self._request("post", "run", json=zones_to_run)



"""Define package errors."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class SprinklError(Exception):
    """Base class for all errors."""

    pass


class InvalidWebhookEvent(Exception):
    """Invalid webhook event."""

    pass


class AuthenticateError(SprinklError):
    """Authentication error."""

    pass


class TokenExpired(SprinklError):
    """Token expired error."""

    pass


class RequestTimeout(SprinklError):
    """Request timeout."""

    pass


class RequestError(SprinklError):
    """Request error."""

    pass


class ControllerAlreadyRunning(SprinklError):
    """Controller already running error."""

    pass



"""Data object helpers."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
from typing import Any, Dict, List


# pylint: disable=inconsistent-return-statements
# List/Dict object contians only know types
def set_default(obj):
    """Make ListObject/DictObject into list/dict for serialization."""
    if isinstance(obj, ListObject):
        # pylint: disable=W0212
        return obj._data

    if isinstance(obj, DictObject):
        # pylint: disable=W0212
        return obj._data


class ListObject:
    """Represent a list with property access to sub-dict/list objects."""

    # pylint: disable=unused-variable
    def __init__(self, data: list):
        """Initialize."""
        self._data = []  # type: List[Any]
        for pos, item in enumerate(data):
            if isinstance(item, dict):
                self._data.append(DictObject(item))
            elif isinstance(item, list):
                self._data.append(ListObject(item))
            else:
                self._data.append(item)

    def __iter__(self):
        """Iterator."""
        for item in self._data:
            yield item

    def __getitem__(self, idx):
        """Iterator."""
        if idx < 0 or idx >= len(self._data):
            raise KeyError

        return self._data[idx]

    def __len__(self):
        """Return number of items."""
        return len(self._data)

    def get(self, idx):
        """Return item based on index."""
        return self._data[idx]

    def __str__(self):
        """Return string (json) representing the object."""
        return self.json

    @property
    def data(self):
        """Return data object."""
        return self._data

    @property
    def json(self) -> str:
        """Return a well-formated json string."""
        return json.dumps(self._data, sort_keys=True, indent=4, default=set_default)


class DictObject:
    """Represent a dict with property access to sub-dict/list objects."""

    def __init__(self, data: dict):
        """Initialize."""
        self._data = {}  # type: Dict[Any, Any]
        for key in data:
            if isinstance(data[key], dict):
                self._data[key] = DictObject(data[key])
            elif isinstance(data[key], list):
                self._data[key] = ListObject(data[key])
            else:
                self._data[key] = data[key]

    def __getattr__(self, name):
        """Allow property name access to data."""
        if name in self._data:
            return self._data[name]

        return object.__getattribute__(self, name)

    def __iter__(self):
        """Iterator."""
        for item in self._data:
            yield item

    def __getitem__(self, key):
        """Iterator."""
        if key in self._data:
            return self._data[key]
        raise KeyError

    def __len__(self):
        """Return number of items."""
        return len(self._data)

    def get(self, key):
        """Return key."""
        return self._data.get(key)

    def __str__(self):
        """Return string (json) representing the object."""
        return self.json

    @property
    def data(self):
        """Return data object."""
        return self._data

    @property
    def json(self) -> str:
        """Return a well-formated json string."""
        return json.dumps(self._data, sort_keys=True, indent=4, default=set_default)



"""Define a class to hold schedule classes."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Awaitable, Callable, Dict

from .schedule import Schedule


class Schedules:
    """Defines the schedules for a controller."""

    def __init__(
        self,
        request: Callable[..., Awaitable[dict]],
        schedules: dict,
        adjustments: dict,
    ) -> None:
        """Intialize."""
        self._adjustments = adjustments
        self._schedules: Dict[str, Any] = {}
        for schedule in schedules:
            parsed_schedule = Schedule(request, adjustments, schedule)
            self._schedules[parsed_schedule.id] = parsed_schedule

    def __iter__(self):
        """Iterator."""
        for item in self._schedules:
            yield self._schedules[item]

    def __getitem__(self, key):
        """Get sensor by id."""
        if key in self._schedules:
            return self._schedules[key]
        raise KeyError

    def __len__(self):
        """Return number of schedules."""
        return len(self._schedules)

    def get(self, key: str):
        """Return sensor by id."""
        return self._schedules.get(key)

    async def all(self, include_disabled: bool = True) -> list:
        """Return all or active schedules."""
        return [
            self._schedules[key]
            for key in self._schedules
            if include_disabled or self._schedules[key].enabled
        ]



"""Define moisture sensors class."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Awaitable, Callable, Dict

from .moisturesensor import MoistureSensor


class MoistureSensors:
    """Class for moisture sensors."""

    def __init__(self, request: Callable[..., Awaitable[dict]], sensors: list) -> None:
        """Intialize."""
        self._request = request
        self._sensors: Dict[str, Any] = {}
        for sensor in sensors:
            parsed_sensor = MoistureSensor(self._request, sensor)
            self._sensors[parsed_sensor.id] = parsed_sensor

    def __iter__(self):
        """Iterator."""
        for item in self._sensors:
            yield self._sensors[item]

    def __getitem__(self, key):
        """Get sensor by id."""
        if key in self._sensors:
            return self._sensors[key]
        raise KeyError

    def __len__(self):
        """Return number of sensors."""
        return len(self._sensors)

    def get(self, key: str):
        """Return sensor by id."""
        return self._sensors.get(key)

    def all(self, include_disabled: bool = True) -> list:
        """Return all or active sensors."""
        return [
            self._sensors[key]
            for key in self._sensors
            if include_disabled or self._sensors[key].enabled
        ]



"""Define moisture sensor class."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Awaitable, Callable

from .pageobject import PageObject


class MoistureSensor:
    """Class for a moisture sensor."""

    def __init__(self, request: Callable[..., Awaitable[dict]], sensor: dict) -> None:
        """Initialize."""
        self._request = request
        self._data = sensor
        # not useful data (no timestamps)
        del self._data["temps"]
        del self._data["moistures_t"]
        del self._data["moistures_m"]
        del self._data["moistures_b"]

        # convert so we are consistent
        self._data["moisture_1"] = self._data["moisture_t"]
        self._data["moisture_3"] = self._data["moisture_m"]
        self._data["moisture_5"] = self._data["moisture_b"]

        del self._data["moisture_t"]
        del self._data["moisture_m"]
        del self._data["moisture_b"]

    def __getattr__(self, name):
        """Allow property access of data object."""
        if name in self._data:
            return self._data[name]

        return object.__getattribute__(self, name)

    async def readings(self) -> PageObject:
        """Return sensor readings."""
        data = await self._request("get", "sensors/{0}/readings".format(self.id))

        return PageObject(
            data, self._request, "get", "sensors/{0}/readings".format(self.id)
        )

    async def averages_day(self) -> PageObject:
        """Return average of day from sensor readings."""
        data = await self._request(
            "get", "sensors/{0}/readings/averages/day".format(self.id)
        )

        return PageObject(
            data,
            self._request,
            "get",
            "sensors/{0}/readings/averages/day".format(self.id),
        )

    async def averages_hour(self) -> PageObject:
        """Return average of hour from sensor readings."""
        data = await self._request(
            "get", "sensors/{0}/readings/averages/hour".format(self.id)
        )

        return PageObject(
            data,
            self._request,
            "get",
            "sensors/{0}/readings/averages/hour".format(self.id),
        )

    async def refresh(self) -> None:
        """Refresh sensor object from Sprinkl controller."""
        data = await self._request("get", "sensors/{0}".format(self.id))
        for key in data["data"]:
            self._data[key] = data["data"][key]



"""Define classes for webhook events."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .dataobject import DictObject


class WebhookBaseObject:
    """Base object for all webhook objects."""

    def __init__(self, obj):
        """Intialize."""
        self._device_id = obj["device_id"]
        self._webhook_id = obj["webhook_id"]
        self._external_id = obj["external_id"]
        self._event = obj["event"]
        self._event_data = DictObject(obj["event_data"])

    @property
    def device_id(self):
        """Return device id."""
        return self._device_id

    @property
    def webhook_id(self):
        """Return webhook id."""
        return self._webhook_id

    @property
    def external_id(self):
        """Return external id."""
        return self._external_id

    @property
    def event(self):
        """Return event type."""
        return self._event

    @property
    def data(self) -> DictObject:
        """Return event data."""
        return self._event_data



"""Controller class for Sprinkl controllers."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from typing import Awaitable, Callable

from .dataobject import DictObject
from .moisturesensors import MoistureSensors
from .pageobject import PageObject
from .schedules import Schedules
from .zones import Zones
from .webhooks import Webhooks

_LOGGER = logging.getLogger(__name__)

URL_BASE = "https://api.sprinkl.com/v1"


class Controller:
    """Controller for a Sprinkl device."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self, request: Callable[..., Awaitable[dict]], controller_data: dict
    ) -> None:
        """Initialize."""
        self._request = request
        self._moisture_sensors = MoistureSensors(
            self._request_controller, controller_data["moisture_sensors"]
        )
        self._weather = DictObject(controller_data["weather"])
        self._schedules = Schedules(
            self._request_controller,
            controller_data["schedules"],
            controller_data["conservation"]["seasonal_adjustments"],
        )
        self._location = DictObject(controller_data["location"])
        self._conservation = DictObject(controller_data["conservation"])
        self._zones = Zones(self._request_controller, controller_data["zones"])
        self._controller = controller_data
        self._webhooks = None

    def __getattr__(self, name):
        """Allow property access to data object."""
        if name in self._controller:
            return self._controller[name]

        return object.__getattribute__(self, name)

    async def _request_controller(
        self,
        method: str,
        path: str,
        *,
        headers: dict = None,
        params: dict = None,
        json: dict = None
    ):
        return await self._request(
            method,
            "{0}/controllers/{1}/{2}".format(URL_BASE, self._controller["id"], path),
            headers=headers,
            params=params,
            json=json,
            reauth_token=True,
        )

    @property
    def enabled(self) -> bool:
        """Return if the controller is enabled."""
        return self._controller["enabled"]

    @property
    def connected(self) -> bool:
        """Return if the controller is connected."""
        return self._controller["connected"]

    @property
    def webhooks(self):
        """Return helper to handle webhooks."""
        if not self._webhooks:
            self._webhooks = Webhooks(self._request_controller)
        return self._webhooks

    async def moisture_sensors(self) -> MoistureSensors:
        """Return the moister sensors."""
        return self._moisture_sensors

    async def weather(self) -> DictObject:
        """Return the weather."""
        return self._weather

    async def location(self) -> DictObject:
        """Return conservation schedule for the current controller."""
        return self._location

    async def conservation(self) -> DictObject:
        """Return conservation schedule."""
        return self._conservation

    async def schedules(self) -> Schedules:
        """Return the schedules."""
        return self._schedules

    async def zones(self) -> Zones:
        """Return the zones."""
        return self._zones

    async def history(self) -> PageObject:
        """Return history of events."""
        data = await self._request_controller("get", "history")
        return PageObject(data, self._request, "get", "history")

    async def stop(self) -> None:
        """Stop/halt the controller."""
        return await self._request_controller("post", "stop")



"""Global test fixtures"""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json

import aresponses
import pytest
import asyncio
import json

from ..const import TEST_HOST, TEST_PORT


@pytest.fixture()
def fail_to_authenticate_refresh_token(event_loop, auth_token, controller_json):
    client = aresponses.ResponsesMockServer(loop=event_loop)

    """Fail refresh token"""
    client.add(
        TEST_HOST, "/v1/authenticate", "POST", aresponses.Response(status=401, text="")
    )

    """Fail autentication twice"""
    client.add(
        TEST_HOST, "/v1/authenticate", "POST", aresponses.Response(status=401, text="")
    )

    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def login_fixture(event_loop, auth_token, controller_json):
    client = aresponses.ResponsesMockServer(loop=event_loop)

    client.add(
        TEST_HOST,
        "/v1/authenticate",
        "post",
        aresponses.Response(
            text=json.dumps(
                {
                    "data": {
                        "token": "login_token",
                        "refresh_token": "login_refresh_token",
                        "user_id": "login_userid",
                    }
                }
            ),
            status=200,
        ),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers",
        "GET",
        aresponses.Response(status=200, text=json.dumps(controller_json)),
    )

    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def controller_login_fixture(event_loop, auth_token, controller_json):
    """Return an aresponses server for an authenticated remote client."""
    client = aresponses.ResponsesMockServer(loop=event_loop)

    client.add(
        TEST_HOST,
        "/v1/authenticate",
        "post",
        aresponses.Response(
            text=json.dumps(
                {
                    "data": {
                        "token": "login_token",
                        "refresh_token": "login_refresh_token",
                        "user_id": "login_userid",
                    }
                }
            ),
            status=200,
        ),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers",
        "GET",
        aresponses.Response(status=200, text=json.dumps(controller_json)),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers/1/history",
        "GET",
        aresponses.Response(
            status=200,
            text=json.dumps(
                {"data": [{"dummy": "history"}], "meta": {"page": 1, "count": 1}}
            ),
        ),
    )
    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def authenticated_refresh_token(
    event_loop, auth_token, controller_json, authenticate_token_expired_json
):
    def handler(request):
        if "Authorization" in request.headers and (
            request.headers["Authorization"] == "refreshed_token"
            or request.headers["Authorization"] == "login_token"
        ):
            return aresponses.Response(status=200, text=json.dumps(controller_json))

        return aresponses.Response(
            status=401, text=json.dumps(authenticate_token_expired_json)
        )

    def login_handler(request):
        # token refresh
        if "refresh_token" in request.query:
            return aresponses.Response(
                status=200,
                text=json.dumps(
                    {
                        "data": {
                            "token": "refreshed_token",  # used in controllers_handler to signal authenticated user
                            "refresh_token": "refreshed_refresh_token",
                            "user_id": "refreshed_userid",
                        }
                    }
                ),
            )

        # login
        return aresponses.Response(
            status=200,
            text=json.dumps(
                {
                    "data": {
                        "token": "login_token",
                        "refresh_token": "login_refresh_token",
                        "user_id": "login_userid",
                    }
                }
            ),
        )

    """Return an aresponses server for an authenticated remote client."""
    client = aresponses.ResponsesMockServer(loop=event_loop)

    client.add(TEST_HOST, "/v1/controllers", "GET", handler)

    client.add(TEST_HOST, "/v1/controllers", "GET", handler)

    client.add(TEST_HOST, "/v1/authenticate", "post", login_handler)

    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def controller_auth_failure_fixture(
    event_loop, auth_token, controller_json, authenticate_token_expired_json
):
    """Return an aresponses server for an authenticated remote client."""
    client = aresponses.ResponsesMockServer(loop=event_loop)

    client.add(
        TEST_HOST, "/v1/controllers", "GET", aresponses.Response(status=401, text="")
    )

    client.add(
        TEST_HOST,
        "/v1/authenticate",
        "post",
        aresponses.Response(
            status=200,
            text=json.dumps(
                {
                    "data": {
                        "token": "login_token",
                        "refresh_token": "login_refresh_token",
                        "user_id": "login_userid",
                    }
                }
            ),
        ),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers",
        "GET",
        aresponses.Response(status=200, text=json.dumps(controller_json)),
    )

    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def login_fail_refresh_token(event_loop, auth_token, controller_json):
    """Return an aresponses server for an authenticated remote client."""
    client = aresponses.ResponsesMockServer(loop=event_loop)

    """Failed to refresh token"""
    client.add(
        TEST_HOST, "/v1/authenticate", "post", aresponses.Response(status=401, text="")
    )

    client.add(
        TEST_HOST,
        "/v1/authenticate",
        "post",
        aresponses.Response(
            status=200,
            text=json.dumps(
                {
                    "data": {
                        "token": "login_token",
                        "refresh_token": "login_refresh_token",
                        "user_id": "login_userid",
                    }
                }
            ),
        ),
    )

    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def moister_sensor_get_day_average(
    event_loop, auth_token, controller_json, controller_sensor_json
):
    """Return an aresponses server for an authenticated remote client."""
    client = aresponses.ResponsesMockServer(loop=event_loop)

    client.add(
        TEST_HOST,
        "/v1/authenticate",
        "post",
        aresponses.Response(
            text=json.dumps(
                {
                    "data": {
                        "token": "login_token",
                        "refresh_token": "login_refresh_token",
                        "user_id": "login_userid",
                    }
                }
            ),
            status=200,
        ),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers",
        "GET",
        aresponses.Response(status=200, text=json.dumps(controller_json)),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers/1/sensors/1",
        "GET",
        aresponses.Response(status=200, text=json.dumps(controller_sensor_json)),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers/1/sensors/1/readings",
        "GET",
        aresponses.Response(
            status=200,
            text=json.dumps(
                {"data": [{"dummy": "no sensor"}], "meta": {"page": "1", "count": 1}}
            ),
        ),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers/1/sensors/1/readings/averages/day",
        "GET",
        aresponses.Response(
            status=200,
            text=json.dumps(
                {"data": [{"dummy": "no sensor"}], "meta": {"page": "1", "count": 1}}
            ),
        ),
    )

    client.add(
        TEST_HOST,
        "/v1/controllers/1/sensors/1/readings/averages/hour",
        "GET",
        aresponses.Response(
            status=200,
            text=json.dumps(
                {"data": [{"dummy": "no sensor"}], "meta": {"page": "1", "count": 1}}
            ),
        ),
    )

    yield client

    event_loop.run_until_complete(event_loop.shutdown_asyncgens())
    event_loop.close()


@pytest.fixture()
def authenticate_token_expired_json():
    return {"errors": [{"detail": "Token has expired"}]}


@pytest.fixture()
def auth_token():
    return {
        "data": {
            "token": "test-token",
            "refresh_token": "refresh-token",
            "user_id": "test-user",
        }
    }


@pytest.fixture()
def controller_sensor_json():
    return {
        "data": {
            "battery": 90,
            "enabled": True,
            "id": "1",
            "last_reading_at": "2019-06-14T03:07:51.837Z",
            "mac": "20:10:00:00:00:00:00:47",
            "moisture_5": 30,
            "moisture_3": 20,
            "moisture_1": 10,
            "moistures_5": [
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                96,
                97,
                97,
                96,
                97,
                97,
                96,
                97,
                96,
                96,
                97,
                97,
                97,
                97,
                96,
                96,
            ],
            "moistures_3": [
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
                97,
            ],
            "moistures_1": [
                53,
                53,
                52,
                52,
                52,
                52,
                52,
                52,
                52,
                51,
                51,
                51,
                51,
                51,
                50,
                50,
                50,
                50,
                50,
                50,
                50,
                49,
                49,
                49,
                49,
                48,
                48,
                48,
                48,
                48,
                48,
                48,
                48,
                48,
                48,
                47,
                47,
                47,
                47,
                47,
                47,
                47,
                47,
                47,
                47,
                47,
                46,
                46,
            ],
            "name": "Test Sensor",
            "rssi": -79,
            "temp": 81.0,
            "temps": [
                68.0,
                68.0,
                68.0,
                68.0,
                68.0,
                68.0,
                68.0,
                70.0,
                72.0,
                73.0,
                75.0,
                77.0,
                79.0,
                81.0,
                82.0,
                82.0,
                84.0,
                84.0,
                86.0,
                86.0,
                86.0,
                86.0,
                88.0,
                88.0,
                88.0,
                90.0,
                90.0,
                90.0,
                90.0,
                90.0,
                91.0,
                90.0,
                91.0,
                91.0,
                91.0,
                91.0,
                90.0,
                90.0,
                90.0,
                88.0,
                88.0,
                86.0,
                86.0,
                84.0,
                84.0,
                84.0,
                82.0,
                81.0,
            ],
        }
    }


@pytest.fixture()
def webhooks_events_json():
    return {
        "data": [
            "OFFLINE",
            "ONLINE",
            "ZONE_STARTED",
            "ZONE_COMPLETED",
            "ZONE_FAULT",
            "WATERING_STOP",
            "SCHEDULE_STARTED",
            "SCHEDULE_COMPLETED",
            "SCHEDULE_SKIPPED",
            "WEATHER_ALERT",
            "WEATHER_ALERT_REMOVED",
        ]
    }


@pytest.fixture()
def webhook_get_id1_json():
    return {
        "data": {
            "created_at": "2019-06-25T19:08:25.880Z",
            "device_id": "d_1",
            "events": ["OFFLINE", "ONLINE"],
            "external_id": "ext_id1",
            "id": "id_1",
            "updated_at": "2019-06-25T19:08:25.880Z",
            "url": "https://webhooks/ext_id1",
        }
    }


@pytest.fixture()
def webhook_list_json():
    return {
        "data": [
            {
                "created_at": "2019-06-25T19:08:25.880Z",
                "device_id": "d_1",
                "events": ["OFFLINE", "ONLINE"],
                "external_id": "ext_id1",
                "id": "id_1",
                "updated_at": "2019-06-25T19:08:25.880Z",
                "url": "https://webhooks/ext_id1",
            },
            {
                "created_at": "2019-06-25T21:57:49.972Z",
                "device_id": "d_1",
                "events": [
                    "OFFLINE",
                    "ONLINE",
                    "ZONE_STARTED",
                    "ZONE_COMPLETED",
                    "ZONE_FAULT",
                    "WATERING_STOP",
                    "SCHEDULE_STARTED",
                    "SCHEDULE_COMPLETED",
                    "SCHEDULE_SKIPPED",
                    "WEATHER_ALERT",
                    "WEATHER_ALERT_REMOVED",
                ],
                "external_id": "ext_id2",
                "id": "id_2",
                "updated_at": "2019-06-25T21:57:49.972Z",
                "url": "https://webhooks/ext_id2",
            },
            {
                "created_at": "2019-06-28T15:02:04.862Z",
                "device_id": "d_1",
                "events": [
                    "OFFLINE",
                    "ONLINE",
                    "ZONE_STARTED",
                    "ZONE_COMPLETED",
                    "ZONE_FAULT",
                    "WATERING_STOP",
                    "SCHEDULE_STARTED",
                    "SCHEDULE_COMPLETED",
                    "SCHEDULE_SKIPPED",
                    "WEATHER_ALERT",
                    "WEATHER_ALERT_REMOVED",
                ],
                "external_id": "ext_id3",
                "id": "id_3",
                "updated_at": "2019-06-28T15:02:04.862Z",
                "url": "https://webhooks/ext_id3",
            },
        ]
    }


@pytest.fixture()
def controller_json():
    return {
        "data": [
            {
                "alert": "None",
                "connected": True,
                "conservation": {
                    "rain_chance": 70.0,
                    "rain_inches_24hr": 0.3,
                    "rain_inches_48hr": 0.6,
                    "rain_inches_4day": 1.5,
                    "rain_inches_7day": 2.0,
                    "seasonal_adjustments": [
                        20,
                        30,
                        40,
                        80,
                        100,
                        100,
                        100,
                        100,
                        100,
                        80,
                        40,
                        20,
                    ],
                    "temp_above": 35.0,
                },
                "created_at": "2018-09-29T23:34:52.153Z",
                "enabled": True,
                "id": "1",
                "last_checkin_at": "2019-06-14T03:13:28.811Z",
                "last_ran_at": "2019-06-13T08:40:01.000Z",
                "location": {
                    "city": "Palo Alto",
                    "country": "United States",
                    "latitude": 10.41754879871175,
                    "longitude": -100.12193391663871,
                    "postal_code": "94200",
                    "state": "California",
                    "street_1": "None",
                    "street_2": "None",
                    "timezone": "America/Los_Angeles",
                },
                "moisture_sensors": [
                    {
                        "battery": 100,
                        "enabled": True,
                        "id": "1",
                        "last_reading_at": "2019-06-14T03:07:51.837Z",
                        "mac": "20:10:00:00:00:00:00:47",
                        "moisture_b": 96,
                        "moisture_m": 97,
                        "moisture_t": 46,
                        "moistures_b": [
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            96,
                            97,
                            97,
                            96,
                            97,
                            97,
                            96,
                            97,
                            96,
                            96,
                            97,
                            97,
                            97,
                            97,
                            96,
                            96,
                        ],
                        "moistures_m": [
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                            97,
                        ],
                        "moistures_t": [
                            53,
                            53,
                            52,
                            52,
                            52,
                            52,
                            52,
                            52,
                            52,
                            51,
                            51,
                            51,
                            51,
                            51,
                            50,
                            50,
                            50,
                            50,
                            50,
                            50,
                            50,
                            49,
                            49,
                            49,
                            49,
                            48,
                            48,
                            48,
                            48,
                            48,
                            48,
                            48,
                            48,
                            48,
                            48,
                            47,
                            47,
                            47,
                            47,
                            47,
                            47,
                            47,
                            47,
                            47,
                            47,
                            47,
                            46,
                            46,
                        ],
                        "name": "Test Sensor",
                        "rssi": -79,
                        "temp": 81.0,
                        "temps": [
                            68.0,
                            68.0,
                            68.0,
                            68.0,
                            68.0,
                            68.0,
                            68.0,
                            70.0,
                            72.0,
                            73.0,
                            75.0,
                            77.0,
                            79.0,
                            81.0,
                            82.0,
                            82.0,
                            84.0,
                            84.0,
                            86.0,
                            86.0,
                            86.0,
                            86.0,
                            88.0,
                            88.0,
                            88.0,
                            90.0,
                            90.0,
                            90.0,
                            90.0,
                            90.0,
                            91.0,
                            90.0,
                            91.0,
                            91.0,
                            91.0,
                            91.0,
                            90.0,
                            90.0,
                            90.0,
                            88.0,
                            88.0,
                            86.0,
                            86.0,
                            84.0,
                            84.0,
                            84.0,
                            82.0,
                            81.0,
                        ],
                    }
                ],
                "name": "Test Site",
                "next_scheduled_at": "2019-06-14T08:00:52.000Z",
                "schedules": [
                    {
                        "created_at": "2018-09-29T23:51:11.500Z",
                        "days": ["S", "T", "Th", "Su"],
                        "enabled": True,
                        "frequency": "odd_days",
                        "id": "s_1",
                        "name": "2 days",
                        "next_run_at": "None",
                        "run_time": "None",
                        "scheduled_days": [],
                        "seasonally_adjust": False,
                        "start_time": "2018-09-30T14:00:15.000Z",
                        "type": "standard",
                        "updated_at": "2018-12-20T05:22:25.858Z",
                        "zones": [{"id": "z_1", "number": 1, "run_time": 4}],
                    }
                ],
                "updated_at": "2019-06-14T03:16:14.255Z",
                "weather": {
                    "accumulations": {
                        "total_2day": 0.0,
                        "total_4day": 0.0,
                        "total_7day": 0.0,
                        "total_today": 0.0,
                    },
                    "conditions": {
                        "pop": 0.0,
                        "raining": False,
                        "temp": 75.0,
                        "temp_high": 79.0,
                        "temp_low": 56.0,
                        "type": "clear_day",
                        "wind": 0.0,
                    },
                    "station": "None",
                },
                "zones": [
                    {
                        "enabled": True,
                        "exposure": "full_sun",
                        "head_dripline_ft": 0,
                        "head_num": 4,
                        "head_type": "spray",
                        "id": "z_1",
                        "moisture_sensor": {
                            "limit_moisture_b": 95,
                            "limit_moisture_m": 95,
                            "limit_moisture_t": 65,
                            "limit_temp": 35,
                            "moisture_sensor_id": "m_1",
                        },
                        "name": "Test Zone",
                        "number": 1,
                        "slope": "flat",
                        "soil_type": "None",
                        "type": "flowerbed",
                    }
                ],
            }
        ],
        "meta": {"count": 1, "page": 1},
    }



"""Define helper object for handling Sprinkl paging."""
#
# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Awaitable, Callable

from .dataobject import DictObject, ListObject


class PageObject:
    """Helper object to handle paging."""

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        paged_data: dict,
        request: Callable[..., Awaitable[dict]],
        method: str,
        uri: str,
        page: int = 1,
    ) -> None:
        """Initialize."""
        self._meta = DictObject(paged_data["meta"])
        self._data = ListObject(paged_data["data"])
        self._page = page

        self._request = request
        self._method = method
        self._uri = uri

    @property
    def has_more(self) -> bool:
        """Return true if there is more pages."""
        if self._page < int(self._meta.count):
            return True
        return False

    async def next(self):
        """Return the next page."""
        if not self.has_more:
            return None

        page = self._page + 1
        data = await self._request(self._method, self._uri, params={"page": page})
        return PageObject(data, self._request, self._method, self._uri, page=page)

    @property
    def meta(self):
        """Return the meta data object."""
        return self._meta

    @property
    def data(self):
        """Return data object."""
        return ListObject(self._data)

    @property
    def json(self) -> str:
        """Return a well-formated json string."""
        return self._data.json
